package util

import (
	"fmt"
	"math"
	"sort"
	"strings"

	"k8s.io/apimachinery/pkg/api/resource"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"

	log "github.com/sirupsen/logrus"
	corev1 "k8s.io/api/core/v1"
	v1 "k8s.io/api/core/v1"

	// Quark
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
)

// StringSetDifference returns the difference between two sets (sets can't have
// duplicates): all the elements in the set a that are not in set b.
func StringSetDifference(a, b []string) (diff []string) {
	m := make(map[string]bool)

	for _, item := range b {
		m[item] = true
	}

	for _, item := range a {
		if !m[item] {
			diff = append(diff, item)
		}
	}
	return
}

// FindStringInSlice returns the index of the first matching string in the slice,
// or -1 if there are no matches.
func FindStringInSlice(slice []string, find string) int {
	for idx, str := range slice {
		if str == find {
			return idx
		}
	}
	return -1
}

func RemoveStringFromSlice(slice []string, s string) (result []string) {
	for _, item := range slice {
		if item == s {
			continue
		}
		result = append(result, item)
	}
	return
}

// StringSetContains returns true if a set includes an item.
func StringSetContains(s []string, str string) bool {
	return (FindStringInSlice(s, str) >= 0)
}

func IsEqualStrings(a1 []string, a2 []string) bool {
	sort.Strings(a1)
	sort.Strings(a2)
	if len(a1) == len(a2) {
		for i, v := range a1 {
			if v != a2[i] {
				return false
			}
		}
	} else {
		return false
	}

	return true
}

func ContainsEqualFoldSubstring(str, substr string) bool {
	x := strings.ToLower(str)
	y := strings.ToLower(substr)
	return strings.Contains(x, y)
}

func GetNetAppVolumeCondition(
	volume *quarkv1alpha1.NetAppVolume,
	conditionType quarkv1alpha1.NetAppVolumeConditionType,
) *quarkv1alpha1.NetAppVolumeCondition {
	var condition *quarkv1alpha1.NetAppVolumeCondition
	for _, cond := range volume.Status.Conditions {
		if cond.Type == conditionType {
			condition = &cond
			break
		}
	}
	return condition
}

func GetNetAppStoragePoolCondition(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	conditionType quarkv1alpha1.NetAppStoragePoolConditionType,
) *quarkv1alpha1.NetAppStoragePoolCondition {
	var condition *quarkv1alpha1.NetAppStoragePoolCondition
	for _, cond := range storagePool.Status.Conditions {
		if cond.Type == conditionType {
			condition = &cond
			break
		}
	}
	return condition
}

func IsSMBConfigurationEqual(volume *quarkv1alpha1.NetAppVolume) bool {
	return (IsEqualSMBActiveDirectoryConfiguration(volume) &&
		IsEqualSMBShareProperties(volume))
}

func IsSMBServerConfigurationEqual(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return (storagePool.Spec.SMBServerConfiguration == storagePool.Status.SMBServerConfiguration)
}

// IsStoragePoolSMBConfigured returns "true" if SMB server id configured in storagepool.
func IsStoragePoolSMBConfigured(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	if storagePool.Spec.SMBServerConfiguration.Netbios != "" && storagePool.Spec.SMBServerConfiguration.Domain != "" {
		return true
	}
	return false
}

func GetVolCreateADSecretName(volumeName string) string {
	return volumeName + "ad-create"
}

func GetVolDeleteADSecretName(volumeName string) string {
	return volumeName + "ad-delete"
}
func GetVolPwdResetADSecretName(volumeName string) string {
	return volumeName + "ad-reset"
}

// GetContainerSpec gets the container spec by containerName.
func GetContainerSpec(containers []corev1.Container, containerName string) *corev1.Container {
	for i, c := range containers {
		if containerName == c.Name {
			return &containers[i]
		}
	}
	return nil
}

func MakeResolvConfString(dnsIPs []string) string {
	resolvData := ""
	for _, dnsIP := range dnsIPs {
		resolvData += "nameserver " + dnsIP + "\n"
	}
	return resolvData
}

func RoundUp(value float64, multiple int64) int64 {
	fm := float64(multiple)
	return int64(math.Ceil(value/fm) * fm)
}

// Allowed characters in kubernetes object names are alphanumeric, "-", and ".".
// K8s zones are of the form "<geo>-<subregion>-<zoneNumber>". Use "." as a
// separator, since the string is used to name storage classes and we need
// to be able to easily undo the join.
func GetFailureDomainString(zones []string) string {
	sort.Strings(zones)
	return strings.Join(zones, ".")
}

func GetFailureDomainZones(fd string) []string {
	return strings.Split(fd, ".")
}

func GetSCNameForFailureDomain(baseName, failureDomain string) string {
	if failureDomain == "" {
		return baseName
	}
	return fmt.Sprintf("%s-%s", baseName, failureDomain)
}

// If a node becomes unreachable, the pod's Ready status will be set to ConditionUnknown,
// but the pod itself will still be in the Running phase. In general, it seems like
// once a pod enters the Running phase, the phase doesn't change. To check if the pod is
// truly running, we check against the Phase and PodReady condition.
func IsPodRunning(pod *corev1.Pod) bool {
	var podReady = false
	for _, cond := range pod.Status.Conditions {
		if cond.Type == corev1.PodReady {
			podReady = cond.Status == corev1.ConditionTrue
			break
		}
	}
	return pod.Status.Phase == corev1.PodRunning && podReady
}

func PrepullContainersFromImageMap(data map[string]string, cloud string) []corev1.Container {
	prepullContainers := []corev1.Container{}
	for k, v := range data {
		container := corev1.Container{
			Name:  k,
			Image: v,
			// available only when shareProcessNamespace field is set
			Command: []string{"/proc/1/root/pause"},
		}
		capability := &v1.Capabilities{Add: []v1.Capability{"SYS_PTRACE"}}
		container.SecurityContext = &v1.SecurityContext{Capabilities: capability}
		prepullContainers = append(prepullContainers, container)
	}
	return prepullContainers
}

func IsEqualSMBShareProperties(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Spec.SMBConfiguration.EncryptionEnabled == volume.Status.SMBConfiguration.EncryptionEnabled &&
		volume.Spec.SMBConfiguration.HideShare == volume.Status.SMBConfiguration.HideShare &&
		volume.Spec.SMBConfiguration.ShareAccessBasedEnumeration == volume.Status.SMBConfiguration.ShareAccessBasedEnumeration)

}

func IsEqualSMBActiveDirectoryConfiguration(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Spec.SMBConfiguration.Netbios == volume.Status.SMBConfiguration.Netbios &&
		volume.Spec.SMBConfiguration.Domain == volume.Status.SMBConfiguration.Domain &&
		volume.Spec.SMBConfiguration.EncryptedADCommunication == volume.Status.SMBConfiguration.EncryptedADCommunication &&
		volume.Spec.SMBConfiguration.PreferredADSite == volume.Status.SMBConfiguration.PreferredADSite &&
		IsEqualStrings(volume.Spec.SMBConfiguration.DNSIPs, volume.Status.SMBConfiguration.DNSIPs))

}

func NodeSelectorMapToString(nodeSelector map[string]string) (out string) {
	var keys []string
	for k := range nodeSelector {
		keys = append(keys, k)
	}
	sort.Strings(keys)
	for _, key := range keys {
		if out == "" {
			out = fmt.Sprintf("%s=%s", key, nodeSelector[key])
		} else {
			out = fmt.Sprintf("%s,%s=%s", out, key, nodeSelector[key])
		}
	}
	return
}

func NodeSelectorStringToMap(nodeSelector string) map[string]string {
	out := make(map[string]string)
	if len(nodeSelector) == 0 {
		return out
	}
	kvPairs := strings.Split(nodeSelector, ",")
	for _, pair := range kvPairs {
		kAndV := strings.Split(pair, "=")
		if len(kAndV) != 2 {
			log.WithFields(log.Fields{
				"nodeSelector": nodeSelector,
				"kAndV":        kAndV,
			}).Error("Invalid k/v pair")
			continue
		}
		out[kAndV[0]] = kAndV[1]
	}
	return out
}

func GetZonesForNodes(nodes []*corev1.Node) []string {
	zoneMap := make(map[string]bool)
	for _, node := range nodes {
		if zone, ok := node.Labels[quarkconfig.ZoneTopologyKey]; ok {
			zoneMap[zone] = true
		}
	}
	zones := make([]string, 0, len(zoneMap))
	for zone := range zoneMap {
		zones = append(zones, zone)
	}
	// This is needed to ensure stability when determining things like zone pairs.
	sort.Strings(zones)
	return zones
}

func GetResources(quarkPodBucket quarkv1alpha1.QuarkPodBucket) corev1.ResourceList {
	cpuRequest := resource.NewQuantity(0, resource.DecimalSI)
	memRequest := resource.NewQuantity(0, resource.BinarySI)
	for _, resources := range quarkPodBucket.ContainerResources {
		cpuRequest.Add(resources.CPUsRequest)
		memRequest.Add(resources.MemRequest)
	}
	return corev1.ResourceList{
		corev1.ResourceCPU:    *cpuRequest,
		corev1.ResourceMemory: *memRequest,
	}
}

// Returns the nearest power of 2 to the number rounded up or down
func NearestPowerofTwo(num float64, roundingfunction string) float64 {
	if roundingfunction == quarkconfig.CeilRoundFunction {
		return math.Pow(2, math.Ceil(math.Log2(num)))
	}
	return math.Pow(2, math.Floor(math.Log2(num)))
}

// Given a range of numbers , return all the numbers in the range as a string slice
func MakeRange(min, max int) []string {
	numList := make([]string, max-min+1)
	for i := range numList {
		numList[i] = fmt.Sprint((min + i))
	}
	return numList
}

func IsVersioned1_0Volume(volume *quarkv1alpha1.NetAppVolume) bool {
	if volume.Status.QubeVersion == "1.0" && volume.Status.Size.Value() != 0 {
		return true
	}
	return false
}

// Obtain QoS limits based on service-level requested
func GetServiceLevelLimits(
	svl string,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (IOPSPerTiB, throughput int, capacityLimit resource.Quantity, diskIOPSPerGiB, iopsLimit, tputLimit int, ampFactor float32, err error) {
	switch quarkconfig.ServiceLevel(svl) {
	case quarkconfig.ServiceLevelStandard, quarkconfig.ServiceLevelRegionalStandard:
		IOPSPerTiB = serviceLevel.Spec.IOPSPerTiB
		diskIOPSPerGiB = serviceLevel.Spec.DiskIOPSPerGiB
		throughput = serviceLevel.Spec.MiBsPerTiB
		capacityLimit = serviceLevel.Spec.CapacityLimit
		ampFactor = serviceLevel.Spec.TpAmpFactor
		iopsLimit = serviceLevel.Spec.IOPSLimit
		tputLimit = serviceLevel.Spec.TputLimit
		return
	}
	err = fmt.Errorf("invalid ServiceLevel %v", serviceLevel)
	return
}
